from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

from pdm.backend.hooks.base import Context

logger = logging.getLogger("hooks")


class BuildHook:
    def __init__(self, name: str = "2") -> None:
        self.name = name

    def pdm_build_clean(self, context: Context) -> None:
        logger.info("Hook%s build clean called", self.name)

    def pdm_build_initialize(self, context: Context) -> None:
        logger.info("Hook%s build initialize called", self.name)

    def pdm_build_update_files(self, context: Context, files: dict[str, Path]) -> None:
        logger.info("Hook%s build update files called", self.name)

    def pdm_build_finalize(self, context: Context, artifact: Path) -> None:
        logger.info("Hook%s build finalize called", self.name)


# Hook instance
hook3 = BuildHook("3")



"""
PEP-517 compliant pyproject build-system API
"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Mapping


def get_requires_for_build_wheel(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    """
    Returns an additional list of requirements for building, as PEP508 strings,
    above and beyond those specified in the pyproject.toml file.

    When C-extension build is needed, setuptools should be required, otherwise
    just return an empty list.
    """
    from pdm.backend.wheel import WheelBuilder

    with WheelBuilder(Path.cwd(), config_settings) as builder:
        if builder.config.build_config.run_setuptools:
            return ["setuptools>=40.8.0"]
        return []


def get_requires_for_build_sdist(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    """There isn't any requirement for building a sdist at this point."""
    return []


def prepare_metadata_for_build_wheel(
    metadata_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Prepare the metadata, places it in metadata_directory"""
    from pdm.backend.wheel import WheelBuilder

    with WheelBuilder(Path.cwd(), config_settings) as builder:
        return builder.prepare_metadata(metadata_directory).name


def build_wheel(
    wheel_directory: str,
    config_settings: Mapping[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    """Builds a wheel, places it in wheel_directory"""
    from pdm.backend.wheel import WheelBuilder

    with WheelBuilder(Path.cwd(), config_settings) as builder:
        return builder.build(
            wheel_directory, metadata_directory=metadata_directory
        ).name


def build_sdist(
    sdist_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Builds an sdist, places it in sdist_directory"""
    from pdm.backend.sdist import SdistBuilder

    with SdistBuilder(Path.cwd(), config_settings) as builder:
        return builder.build(sdist_directory).name


def get_requires_for_build_editable(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    """
    Returns an additional list of requirements for building, as PEP508 strings,
    above and beyond those specified in the pyproject.toml file.

    When C-extension build is needed, setuptools should be required, otherwise
    just return an empty list.
    """
    return get_requires_for_build_wheel(config_settings) + ["editables"]


def prepare_metadata_for_build_editable(
    metadata_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Prepare the metadata, places it in metadata_directory"""
    from pdm.backend.editable import EditableBuilder

    with EditableBuilder(Path.cwd(), config_settings) as builder:
        return builder.prepare_metadata(metadata_directory).name


def build_editable(
    wheel_directory: str,
    config_settings: Mapping[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    from pdm.backend.editable import EditableBuilder

    with EditableBuilder(Path.cwd(), config_settings) as builder:
        return builder.build(
            wheel_directory, metadata_directory=metadata_directory
        ).name



import contextlib
import tarfile
import zipfile
from pathlib import Path
from typing import Iterator, List

from pdm.backend import utils
from tests import FIXTURES


def get_tarball_names(path: Path) -> List[str]:
    with tarfile.open(path, "r:gz") as tar:
        return tar.getnames()


def get_wheel_names(path: Path) -> List[str]:
    with zipfile.ZipFile(path) as zf:
        return zf.namelist()


@contextlib.contextmanager
def build_fixture_project(project_name: str) -> Iterator[Path]:
    project = FIXTURES / "projects" / project_name
    with utils.cd(project):
        yield project



"""A built-in hook to generate setup.py and run the script"""
from __future__ import annotations

import atexit
import pickle
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import cast

from pdm.backend.exceptions import BuildError
from pdm.backend.hooks.base import Context

# A minimal template of setup.py, which is used to build extensions
SETUP_FORMAT = """\
# -*- coding: utf-8 -*-
from setuptools import setup

{before}
setup_kwargs = {{
    'name': {name!r},
    'version': {version!r},
    'description': {description!r},
    'url': {url!r},
{extra}
}}
{after}

setup(**setup_kwargs)
"""

HOOK_TEMPLATE = """\
import pickle

context_dump = {context_dump!r}
context = pickle.loads(context_dump)
builder = context.builder
builder.call_hook("pdm_build_update_setup_kwargs", context, setup_kwargs)
"""


def _format_list(data: list[str], indent: int = 4) -> str:
    result = ["["]
    for row in data:
        result.append(" " * indent + repr(row) + ",")
    result.append(" " * (indent - 4) + "]")
    return "\n".join(result)


def _format_dict_list(data: dict[str, list[str]], indent: int = 4) -> str:
    result = ["{"]
    for key, value in data.items():
        result.append(
            " " * indent + repr(key) + ": " + _format_list(value, indent + 4) + ","
        )
    result.append(" " * (indent - 4) + "}")
    return "\n".join(result)


class SetuptoolsBuildHook:
    """A build hook to run setuptools build command."""

    def pdm_build_hook_enabled(self, context: Context) -> bool:
        return context.target != "sdist" and context.config.build_config.run_setuptools

    def pdm_build_update_files(self, context: Context, files: dict[str, Path]) -> None:
        if context.target == "editable":
            self._build_inplace(context)
        else:
            self._build_lib(context)

    def _build_lib(self, context: Context) -> None:
        build_dir = context.ensure_build_dir()
        setup_py = self.ensure_setup_py(context)
        with tempfile.TemporaryDirectory(prefix="pdm-build-") as temp_dir:
            build_args = [sys.executable, str(setup_py), "build", "-b", temp_dir]
            try:
                subprocess.check_call(build_args)
            except subprocess.CalledProcessError as e:
                raise BuildError(f"Error occurs when running {build_args}:\n{e}")
            lib_dir = next(Path(temp_dir).glob("lib.*"), None)
            if not lib_dir:
                return
            # copy the files under temp_dir/lib.* to context.build_dir
            for file in lib_dir.iterdir():
                if file.is_dir():
                    shutil.copytree(file, build_dir / file.name)
                else:
                    shutil.copy2(file, build_dir)

    def _build_inplace(self, context: Context) -> None:
        setup_py = self.ensure_setup_py(context)
        build_args = [sys.executable, str(setup_py), "build_ext", "--inplace"]
        try:
            subprocess.check_call(build_args)
        except subprocess.CalledProcessError as e:
            raise BuildError(f"Error occurs when running {build_args}:\n{e}")

    def ensure_setup_py(self, context: Context, clean: bool = True) -> Path:
        """Ensures the requirement has a setup.py ready."""
        # XXX: Currently only handle PDM project, and do nothing if not.

        setup_py_path = context.root.joinpath("setup.py")
        if setup_py_path.is_file():
            return setup_py_path

        setup_py_path.write_text(self.format_setup_py(context), encoding="utf-8")

        # Clean this temp file when process exits
        def cleanup() -> None:
            try:
                setup_py_path.unlink()
            except OSError:
                pass

        if clean:
            atexit.register(cleanup)
        return setup_py_path

    def format_setup_py(self, context: Context) -> str:
        before, extra, after = [], [], []
        meta = context.config.metadata
        kwargs = {
            "name": meta["name"],
            "version": meta.get("version", "0.0.0"),
            "description": meta.get("description", "UNKNOWN"),
            "url": (meta.get("project-urls", {})).get("homepage", ""),
        }

        # Run the pdm_build_update_setup_kwargs hook to update the kwargs
        after.append(HOOK_TEMPLATE.format(context_dump=pickle.dumps(context)))

        package_paths = context.config.convert_package_paths()
        if package_paths["packages"]:
            extra.append(
                "    'packages': {},\n".format(
                    _format_list(cast("list[str]", package_paths["packages"]), 8)
                )
            )
        if package_paths["package_dir"]:
            extra.append(
                "    'package_dir': {!r},\n".format(package_paths["package_dir"])
            )
        if package_paths["package_data"]:
            extra.append(
                "    'package_data': {!r},\n".format(package_paths["package_data"])
            )
        if package_paths["exclude_package_data"]:
            extra.append(
                "    'exclude_package_data': {!r},\n".format(
                    package_paths["exclude_package_data"]
                )
            )

        if meta.get("dependencies"):
            before.append(f"INSTALL_REQUIRES = {_format_list(meta['dependencies'])}\n")
            extra.append("    'install_requires': INSTALL_REQUIRES,\n")
        if meta.get("optional-dependencies"):
            before.append(
                "EXTRAS_REQUIRE = {}\n".format(
                    _format_dict_list(meta["optional-dependencies"])
                )
            )
            extra.append("    'extras_require': EXTRAS_REQUIRE,\n")
        if meta.get("requires-python"):
            extra.append(f"    'python_requires': {meta['requires-python']!r},\n")
        entry_points = meta.entry_points
        if entry_points:
            entry_points_list = {
                group: [f"{k} = {v}" for k, v in values.items()]
                for group, values in entry_points.items()
            }
            before.append(f"ENTRY_POINTS = {_format_dict_list(entry_points_list)}\n")
            extra.append("    'entry_points': ENTRY_POINTS,\n")
        return SETUP_FORMAT.format(
            before="".join(before), after="".join(after), extra="".join(extra), **kwargs
        )



"""
module to get version from tag of SCM repository.
Adapted from setuptools-scm. Currently only support git and hg.
"""
from __future__ import annotations

import os
import re
import shlex
import shutil
import subprocess
import warnings
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, NamedTuple

from pdm.backend._vendor.packaging.version import Version

if TYPE_CHECKING:
    from _typeshed import StrPath

DEFAULT_TAG_REGEX = re.compile(
    r"^(?:[\w-]+-)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]*)(?:\+.*)?$"
)


@dataclass(frozen=True)
class Config:
    tag_regex: re.Pattern


def _subprocess_call(
    cmd: str | list[str],
    cwd: StrPath | None = None,
    extra_env: dict[str, str] | None = None,
) -> tuple[int, str, str]:
    # adapted from pre-commit
    # Too many bugs dealing with environment variables and GIT:
    # https://github.com/pre-commit/pre-commit/issues/300
    env = {
        k: v
        for k, v in os.environ.items()
        if not k.startswith("GIT_")
        or k in ("GIT_EXEC_PATH", "GIT_SSH", "GIT_SSH_COMMAND")
    }
    env.update({"LC_ALL": "C", "LANG": "", "HGPLAIN": "1"})
    if extra_env:
        env.update(extra_env)
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    proc = subprocess.Popen(
        cmd, cwd=cwd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = proc.communicate()
    return (
        proc.returncode,
        out.decode("utf-8", "surrogateescape").strip(),
        err.decode("utf-8", "surrogateescape").strip(),
    )


class VersionInfo(NamedTuple):
    version: Version
    distance: int | None
    dirty: bool
    node: str | None
    branch: str | None


def meta(
    config: Config,
    tag: str | Version,
    distance: int | None = None,
    dirty: bool = False,
    node: str | None = None,
    branch: str | None = None,
) -> VersionInfo:
    if isinstance(tag, str):
        tag = tag_to_version(config, tag)
    return VersionInfo(tag, distance, dirty, node, branch)


def _git_get_branch(root: StrPath) -> str | None:
    ret, out, _ = _subprocess_call("git rev-parse --abbrev-ref HEAD", root)
    if not ret:
        return out
    return None


def _git_is_dirty(root: StrPath) -> bool:
    _, out, _ = _subprocess_call("git status --porcelain --untracked-files=no", root)
    return bool(out)


def _git_get_node(root: StrPath) -> str | None:
    ret, out, _ = _subprocess_call("git rev-parse --verify --quiet HEAD", root)
    if not ret:
        return out[:7]
    return None


def _git_count_all_nodes(root: StrPath) -> int:
    _, out, _ = _subprocess_call("git rev-list HEAD", root)
    return out.count("\n") + 1


def _git_parse_describe(describe_output: str) -> tuple[str, int, str, bool]:
    # 'describe_output' looks e.g. like 'v1.5.0-0-g4060507' or
    # 'v1.15.1rc1-37-g9bd1298-dirty'.

    if describe_output.endswith("-dirty"):
        dirty = True
        describe_output = describe_output[:-6]
    else:
        dirty = False

    tag, number, node = describe_output.rsplit("-", 2)
    return tag, int(number), node, dirty


class _ParseResult(NamedTuple):
    version: str
    prefix: str
    suffix: str


def _parse_version_tag(config: Config, tag: str) -> _ParseResult | None:
    tagstring = tag if not isinstance(tag, str) else str(tag)
    match = config.tag_regex.match(tagstring)

    result = None
    if match:
        if len(match.groups()) == 1:
            key: int | str = 1
        else:
            key = "version"

        result = _ParseResult(
            match.group(key),
            match.group(0)[: match.start(key)],
            match.group(0)[match.end(key) :],
        )

    return result


def tag_to_version(config: Config, tag: str) -> Version:
    """
    take a tag that might be prefixed with a keyword and return only the version part
    :param config: optional configuration object
    """
    tagdict = _parse_version_tag(config, tag)
    if not tagdict or not tagdict.version:
        warnings.warn(f"tag {tag!r} no version found")
        return Version("0.0.0")

    version = tagdict.version

    if tagdict.suffix:
        warnings.warn(f"tag {tag!r} will be stripped of its suffix {tagdict.suffix!r}")

    return Version(version)


def tags_to_versions(config: Config, tags: Iterable[str]) -> list[Version]:
    """
    take tags that might be prefixed with a keyword and return only the version part
    :param tags: an iterable of tags
    :param config: optional configuration object
    """
    return [tag_to_version(config, tag) for tag in tags if tag]


def git_parse_version(root: StrPath, config: Config) -> VersionInfo | None:
    GIT = shutil.which("git")
    if not GIT:
        return None

    ret, repo, _ = _subprocess_call([GIT, "rev-parse", "--show-toplevel"], root)
    if ret or not repo:
        return None

    if os.path.isfile(os.path.join(repo, ".git/shallow")):
        warnings.warn(f"{repo!r} is shallow and may cause errors")
    describe_cmd = [GIT, "describe", "--dirty", "--tags", "--long", "--match", "*.*"]
    ret, output, err = _subprocess_call(describe_cmd, repo)
    branch = _git_get_branch(repo)

    if ret:
        rev_node = _git_get_node(repo)
        dirty = _git_is_dirty(repo)
        if rev_node is None:
            return meta(config, "0.0", 0, dirty)
        return meta(
            config, "0.0", _git_count_all_nodes(repo), dirty, f"g{rev_node}", branch
        )
    else:
        tag, number, node, dirty = _git_parse_describe(output)
        return meta(config, tag, number or None, dirty, node, branch)


def get_latest_normalizable_tag(root: StrPath) -> str:
    # Gets all tags containing a '.' from oldest to newest
    cmd = [
        "hg",
        "log",
        "-r",
        "ancestors(.) and tag('re:\\.')",
        "--template",
        "{tags}\n",
    ]
    _, output, _ = _subprocess_call(cmd, root)
    outlines = output.split()
    if not outlines:
        return "null"
    tag = outlines[-1].split()[-1]
    return tag


def hg_get_graph_distance(root: StrPath, rev1: str, rev2: str = ".") -> int:
    cmd = ["hg", "log", "-q", "-r", f"{rev1}::{rev2}"]
    _, out, _ = _subprocess_call(cmd, root)
    return len(out.strip().splitlines()) - 1


def _hg_tagdist_normalize_tagcommit(
    config: Config, root: StrPath, tag: str, dist: int, node: str, branch: str
) -> VersionInfo:
    dirty = node.endswith("+")
    node = "h" + node.strip("+")

    # Detect changes since the specified tag
    revset = (
        "(branch(.)"  # look for revisions in this branch only
        " and tag({tag!r})::."  # after the last tag
        # ignore commits that only modify .hgtags and nothing else:
        " and (merge() or file('re:^(?!\\.hgtags).*$'))"
        " and not tag({tag!r}))"  # ignore the tagged commit itself
    ).format(tag=tag)
    if tag != "0.0":
        _, commits, _ = _subprocess_call(
            ["hg", "log", "-r", revset, "--template", "{node|short}"],
            root,
        )
    else:
        commits = "True"

    if commits or dirty:
        return meta(config, tag, distance=dist, node=node, dirty=dirty, branch=branch)
    else:
        return meta(config, tag)


def guess_next_version(tag_version: Version) -> str:
    version = _strip_local(str(tag_version))
    return _bump_dev(version) or _bump_regex(version)


def _strip_local(version_string: str) -> str:
    public, _, _ = version_string.partition("+")
    return public


def _bump_dev(version: str) -> str:
    if ".dev" not in version:
        return ""

    prefix, tail = version.rsplit(".dev", 1)
    assert tail == "0", "own dev numbers are unsupported"
    return prefix


def _bump_regex(version: str) -> str:
    match = re.match(r"(.*?)(\d+)$", version)
    assert match is not None
    prefix, tail = match.groups()
    return "%s%d" % (prefix, int(tail) + 1)


def hg_parse_version(root: StrPath, config: Config) -> VersionInfo | None:
    if not shutil.which("hg"):
        return None
    _, output, _ = _subprocess_call("hg id -i -b -t", root)
    identity_data = output.split()
    if not identity_data:
        return None
    node = identity_data.pop(0)
    branch = identity_data.pop(0)
    if "tip" in identity_data:
        # tip is not a real tag
        identity_data.remove("tip")
    tags = tags_to_versions(config, identity_data)
    dirty = node[-1] == "+"
    if tags:
        return meta(config, tags[0], dirty=dirty, branch=branch)

    if node.strip("+") == "0" * 12:
        return meta(config, "0.0", dirty=dirty, branch=branch)

    try:
        tag = get_latest_normalizable_tag(root)
        dist = hg_get_graph_distance(root, tag)
        if tag == "null":
            tag = "0.0"
            dist = int(dist) + 1
        return _hg_tagdist_normalize_tagcommit(config, root, tag, dist, node, branch)
    except ValueError:
        return None  # unpacking failed, old hg


def format_version(version: VersionInfo) -> str:
    if version.distance is None:
        main_version = str(version.version)
    else:
        guessed = guess_next_version(version.version)
        main_version = f"{guessed}.dev{version.distance}"

    if version.distance is None or version.node is None:
        clean_format = ""
        dirty_format = "+d{time:%Y%m%d}"
    else:
        clean_format = "+{node}"
        dirty_format = "+{node}.d{time:%Y%m%d}"
    fmt = dirty_format if version.dirty else clean_format
    local_version = fmt.format(node=version.node, time=datetime.utcnow())
    return main_version + local_version


def get_version_from_scm(root: str | Path, *, tag_regex: str | None = None) -> str:
    config = Config(tag_regex=re.compile(tag_regex) if tag_regex else DEFAULT_TAG_REGEX)
    for func in (git_parse_version, hg_parse_version):
        version = func(root, config)  # type: ignore
        if version:
            return format_version(version)
    raise ValueError(
        "Cannot find the version from SCM or SCM isn't detected. \n"
        "You can still specify the version via environment variable "
        "`PDM_BUILD_SCM_VERSION`."
    )



from __future__ import annotations

from pathlib import Path

import pytest

from pdm.backend.base import Builder, is_same_or_descendant_path
from pdm.backend.sdist import SdistBuilder
from pdm.backend.wheel import WheelBuilder
from tests import FIXTURES


@pytest.mark.parametrize("builder_cls", (WheelBuilder, SdistBuilder))
def test_auto_include_tests_for_sdist(
    builder_cls: type[Builder], tmp_path: Path
) -> None:
    with builder_cls(FIXTURES / "projects/demo-package-with-tests") as builder:
        context = builder.build_context(tmp_path)
        builder.clean(context)
        builder.initialize(context)
        files = dict(builder.get_files(context))

    sdist_only_files = ("tests/__init__.py", "pyproject.toml")
    include_files = ("my_package/__init__.py",)
    for file in include_files:
        assert file in files

    for file in sdist_only_files:
        if isinstance(builder, SdistBuilder):
            assert file in files
        else:
            assert file not in files


@pytest.mark.parametrize(
    "target,path,expect",
    [
        ("a/b", "a", True),
        ("a/b/c", "a/b/c", True),
        ("b/c", "a", False),
        ("a", "a/b", False),
        ("a", "b/c", False),
    ],
)
def test_is_same_or_descendant_path(target, path, expect) -> None:
    assert is_same_or_descendant_path(target, path) == expect


@pytest.mark.parametrize("builder_cls", (WheelBuilder, SdistBuilder))
def test_recursive_glob_patterns_in_includes(
    builder_cls: type[Builder], tmp_path: Path
) -> None:
    with builder_cls(FIXTURES / "projects/demo-package-with-deep-path") as builder:
        context = builder.build_context(tmp_path)
        builder.clean(context)
        builder.initialize(context)
        files = dict(builder.get_files(context))

    data_files = (
        "my_package/data/data_a.json",
        "my_package/data/data_inner/data_b.json",
    )

    assert "my_package/__init__.py" in files

    for file in data_files:
        if isinstance(builder, WheelBuilder):
            assert file not in files
        else:
            assert file in files


@pytest.mark.parametrize(
    ["includes", "excludes", "data_a_exist", "data_b_exist"],
    [
        (["**/*.json"], ["my_package/data/*.json"], False, True),
        (["my_package/data/data_a.json"], ["my_package/data/*.json"], True, False),
        (
            ["my_package/", "my_package/data/data_a.json"],
            ["my_package/data/data_a.json"],
            False,
            True,
        ),
        (["my_package/data/*"], ["my_package/data/"], True, False),
        (["**/data/*.json"], ["my_package/data/*.json"], False, False),
    ],
)
def test_merge_includes_and_excludes(
    includes, excludes, data_a_exist: bool, data_b_exist: bool, tmp_path: Path
) -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-package-with-deep-path")
    data_a = "my_package/data/data_a.json"
    data_b = "my_package/data/data_inner/data_b.json"

    with builder:
        context = builder.build_context(tmp_path)
        builder.clean(context)
        builder.initialize(context)
        builder.config.build_config["includes"] = includes
        builder.config.build_config["excludes"] = excludes
        builder.config.build_config["source-includes"] = []
        include_files = dict(builder.get_files(context))
        assert (data_a in include_files) == data_a_exist
        assert (data_b in include_files) == data_b_exist


def test_license_file_globs_no_matching() -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-no-license")
    with builder:
        with pytest.warns(UserWarning) as warns:
            license_files = builder.find_license_files()

    assert not license_files
    assert len(warns) == 1
    assert str(warns.pop(UserWarning).message).startswith(
        "No license files are matched with glob patterns"
    )


def test_license_file_paths_no_matching() -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-no-license")
    builder.config.metadata["license-files"] = {"paths": ["LICENSE"]}
    with builder:
        with pytest.raises(ValueError, match="License files not found"):
            builder.find_license_files()


@pytest.mark.parametrize("key", ["paths", "globs"])
def test_license_file_explicit_empty(recwarn, key) -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-no-license")
    builder.config.metadata["license-files"] = {key: []}
    with builder:
        license_files = builder.find_license_files()
    assert not license_files
    assert len(recwarn) == 0


def test_reuse_spec_licenses_dir() -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-reuse-spec")
    with builder:
        license_files = builder.find_license_files()
    assert license_files == ["LICENSES/MPL-2.0.txt"]


def test_collect_build_files_with_src_layout(tmp_path) -> None:
    builder = WheelBuilder(FIXTURES / "projects/demo-src-package")
    with builder:
        context = builder.build_context(tmp_path)
        builder.clean(context)
        builder.initialize(context)
        build_dir = context.ensure_build_dir()
        (build_dir / "my_package").mkdir()
        (build_dir / "my_package" / "hello.py").write_text("print('hello')\n")
        files = dict(builder.get_files(context))
        assert "my_package/hello.py" in files
